!----------------------------------------------------------
!   resolution de l'equation de convection lineaire 2d 
!      (sous forme de loi de conservation d/dt f + div(v f)=0
!   - a champ de vitesse variable
!   - sur grille structure curviligne
!   - domaine topologiquement equivalent a un carre : 
!             NORD
!            _______  
!            |     |  ^j
!    OUEST   |     |  |     EST
!            _______  --> i

!              SUD
!   - schema volumes finis, decentre d'ordre 1, euler explicite en temps
!   - fichiers de donnees : maillage.data
!----------------------------------------------------------

program conv2d_curv

  !----------- sous programmes ------------------------------------ 
  use mod_precision   !- (double ou simple precision)
  use mod_vitesse     !- (champ de vitesse)
  use mod_ci          !- (donnee initiale)
  use mod_aplusmoins  !- (calcul de a+ et a-)
  use mod_sorties     !- resultats de calculs
  !----------------------------------------------------------------

  implicit none

  !------------ variables -----------------------------------------
  integer :: i,j,imax,jmax,nmax,n,k,pas_affichage
  character(len=100) :: numero
  real(pr) :: L,h,dx,dy,cfl,invdt,dt,vmax,tmax
  real(pr), dimension(:,:), allocatable :: x,y,xc,yc
  real(pr), dimension(:,:), allocatable :: aire,v_tau,v_nu,taux,tauy,nux,nuy,flux_tau,flux_nu
  real(pr), dimension(:,:), allocatable :: f,fnpun 
  !----------------------------------------------------------------


  !------------- programme ----------------------------------------

  !--- lecture du maillage
  open(unit=10,file='maillage.data')

  !-nb de points dans la grille
  read(10,*) imax,jmax   

  !-nb de mailles
  imax=imax-1
  jmax=jmax-1   

  allocate(x(0:imax,0:jmax),y(0:imax,0:jmax))
  do i=0,imax
     do j=0,jmax
        read(10,*) x(i,j),y(i,j)
     enddo
  enddo
  close(10)

  !--- coordonnees des centres de mailles
  allocate(xc(1:imax,1:jmax),yc(1:imax,1:jmax))
  do i=1,imax
     do j=1,jmax
        xc(i,j)=0.25_pr*(x(i-1,j-1)+x(i,j-1)+x(i,j)+x(i-1,j))
        yc(i,j)=0.25_pr*(y(i-1,j-1)+y(i,j-1)+y(i,j)+y(i-1,j))
     enddo
  enddo
  
  !--- calcul des aires des mailles
  allocate(aire(1:imax,1:jmax))
  do i=1,imax
     do j=1,jmax
        aire(i,j)=0.5_pr*( (x(i,j)-x(i-1,j-1))*(y(i-1,j)-y(i,j-1))             &
             &            -(y(i,j)-y(i-1,j-1))*(x(i-1,j)-x(i,j-1)) )
     enddo
  enddo

  !--- calcul des normales aux interfaces
  allocate(taux(0:imax,1:jmax),tauy(0:imax,1:jmax))
  allocate(nux(1:imax,0:jmax),nuy(1:imax,0:jmax))
  do i=0,imax
     do j=1,jmax
        taux(i,j) = y(i,j)-y(i,j-1)
        tauy(i,j) = -(x(i,j)-x(i,j-1))
     end do
  end do
  do i=1,imax
     do j=0,jmax
        nux(i,j) = -(y(i,j)-y(i-1,j))
        nuy(i,j) = x(i,j)-x(i-1,j)
     end do
  end do

  !--- donne initiale : pour changer de donnee, modifier le module mod_ci
  allocate(f(0:imax+1,0:jmax+1),fnpun(1:imax,1:jmax))
  f(1:imax,1:jmax)=ci(xc,yc)

  !--- conditions aux limites : 
  !-- exemples 1 et 2 : on met 0 partout sauf E et O : periodique
  f(0,:)=f(imax,:)      !- ouest
  f(imax+1,:)=f(1,:)    !- est
  f(:,0)=0._pr          !- sud
  f(:,jmax+1)=0._pr     !- nord
!!$  !-- exemples 3 et 4 : symtrie sur O et E, 0 ailleurs
!!$  f(0,:)=f(1,:)      !- ouest
!!$  f(imax+1,:)=f(imax,:)    !- est
!!$  f(:,0)=0._pr          !- sud
!!$  f(:,jmax+1)=0._pr     !- nord
!!$  !-- exemple 5 : entre par une partie du bord nord gauche, symtrie en O et E
!!$  f(0,:)=f(1,:)      !- ouest
!!$  f(imax+1,:)=f(imax,:)    !- est
!!$  f(:,0)=0._pr          !- sud
!!$  f(:,jmax+1)=0._pr     !- nord
!!$  f(88:98,jmax+1)=1._pr     !- nord (entre)


  !--- calcul de la "vitesse" normale aux interfaces de mailles
  allocate(v_tau(0:imax,1:jmax),v_nu(1:imax,0:jmax))
  do i=0,imax
     do j=1,jmax
        v_tau(i,j) = vx(0.5_pr*(x(i,j)+x(i,j-1)),0.5_pr*(y(i,j)+y(i,j-1)))*taux(i,j)          &
             &       + vy(0.5_pr*(x(i,j)+x(i,j-1)),0.5_pr*(y(i,j)+y(i,j-1)))*tauy(i,j)          
     end do
  end do
  do i=1,imax
     do j=0,jmax
        v_nu(i,j) = vx(0.5_pr*(x(i,j)+x(i-1,j)),0.5_pr*(y(i,j)+y(i-1,j)))*nux(i,j)          &
             &      + vy(0.5_pr*(x(i,j)+x(i-1,j)),0.5_pr*(y(i,j)+y(i-1,j)))*nuy(i,j)          
     end do
  end do

  !--- pas de temps et vitesse maximum
  cfl=0.9_pr
  invdt=0._pr 
  vmax=0._pr
  do i=1,imax
     do j=1,jmax
        invdt=max(invdt,(abs(v_tau(i,j))+abs(v_nu(i,j)))/aire(i,j))
        vmax=max(vmax,sqrt((vx(x(i,j),y(i,j)))**2+(vy(x(i,j),y(i,j)))**2))
     end do
  end do
  dt=cfl/invdt

  !--- allocation des flux
  allocate(flux_tau(0:imax,1:jmax),flux_nu(1:imax,0:jmax))

  !--- temps de simulation et parametres de sorties
  !-- exemple 1 : rotation rigide
  tmax=4._pr*acos(-1.0_pr) !- (pour faire deux tours)
  nmax=floor(tmax/dt) 
!!$  !-- exemple 2 : couette cylindrique
!!$  tmax=3._pr*acos(-1.0_pr) 
!!$  nmax=floor(tmax/dt) 
!!$  !-- exemples 3, 4 et 5 : champ de vitesse a potentiel complexe autour d'un cylindre (1/2 cylindre)
!!$  tmax=20._pr/vmax
!!$  nmax=floor(tmax/dt) 

  print*,"nombre d'iterations en temps :",nmax
  pas_affichage=max(nmax/120,1)  !-- pour avoir 120 images, soit 5 secondes d'animation
  k=0
  
  !--- evolution
  do n=1,nmax

     !-- calcul des flux
     do i=0,imax
        do j=1,jmax
           flux_tau(i,j) = aplus(v_tau(i,j))*f(i,j) + amoins(v_tau(i,j))*f(i+1,j)
        end do
     end do

     do i=1,imax
        do j=0,jmax
           flux_nu(i,j) = aplus(v_nu(i,j))*f(i,j) + amoins(v_nu(i,j))*f(i,j+1)
        end do
     end do
     

     !-- iteration en temps
     do i=1,imax
        do j=1,jmax
           fnpun(i,j) = f(i,j) - dt/aire(i,j)*( (flux_tau(i,j)-flux_tau(i-1,j)) + (flux_nu(i,j)-flux_nu(i,j-1)) )
        end do
     end do

     !-- mise a jour
     f(1:imax,1:jmax)=fnpun

     !-- mise a jour cl
     !-- exemples 1 et 2 : periodique E et O
     f(0,:)=f(imax,:)      !- est
     f(imax+1,:)=f(1,:)    !- ouest
!!$     !-- exemples 3, 4 et 5 : symtrie sur O et E
!!$     f(0,:)=f(1,:)      !- ouest
!!$     f(imax+1,:)=f(imax,:)    !- est

     !-- sauvegarde resultats (pour medit ou visit)
     if (mod(n,pas_affichage)==0) then

        k=k+1
        write(numero,*) k
        if (k<=9) then
           numero='00'//trim(adjustl(numero))
        elseif (9<k.and.k<=99) then
           numero='0'//trim(adjustl(numero))
        end if

!!$        call sortie_medit(numero,imax,jmax,f,x,y)
        call sortie_visit(numero,imax,jmax,f,x,y)
        
        print*,n
 
     end if

  end do

  !----------------------------------------------------------------

end program conv2d_curv
